from __future__ import print_function
import os, logging, shelve
from itertools import combinations
from qgis.core import QgsVectorLayer, QgsCoordinateReferenceSystem, QgsFeatureRequest


class StructureValidator(object):

	def __init__(self, path, version=4, _format="shp"):
		super(StructureValidator, self).__init__()

		self.__path = path
		self.__version = version
		self.__format = "shapefile" if _format == "shp" else "geodatabase"
		self.__dataStructure = self.get_data_structure()

	@property
	def path(self):
		return self.__path

	def setPath(self, path):
		self.__path = path

	@property
	def version(self):
		return self.__version

	@property
	def _format(self):
		return self.__format

	@property
	def dataStructure(self):
		return self.__dataStructure

	def get_data_structure(self):
		rootDir = os.path.join(os.getcwd(), "template_data")
		#rootDir = os.path.join(os.path.realpath(__file__), "template_data")
		for single in os.listdir(rootDir):
			print(single)
			if os.path.isdir(os.path.join(rootDir,single)):
				ver, form = single.split("_")
				print(ver, form)
				if ver == "v%d" % self.version and form == self._format:
					dataStructure = []
					for root, dirs, files in os.walk(os.path.join(rootDir, single)):
						for singleFile in files:
							name = singleFile.lower()
							if name.endswith(".shp") or name.endswith(".mdb"):
								dataStructure.append(os.path.join(root, singleFile).replace(os.path.join(rootDir, single), ""))
					return dataStructure

	def check_data_structure(self):
		dataStructure = self.dataStructure[:]
		for i in dataStructure:
			checkPath = self.path + i
			check = False
			if os.path.isfile(checkPath):
				check = True
			_type = (os.path.basename(checkPath)).split('.')[-1]
			yield {"check": check, "type": _type, "path": checkPath}


class ShapefileValidator(object):

	def __init__(self, shapefile, crs=32633):
		super(ShapefileValidator, self).__init__()

		self.__name = os.path.basename(shapefile).split('.')[0]
		self.__shapefile = QgsVectorLayer(shapefile, 'shapefile', 'ogr')
		self.__crs = QgsCoordinateReferenceSystem(crs)

		d = shelve.open(os.path.join(os.getcwd(),"cfg","v4_data"))
		self.__fieldsStructure = d["shpFields"][self.name]
		d.close()

	@property
	def name(self):
		return self.__name

	@property
	def crs(self):
		return self.__crs

	def setCrs(self, value):
		self.__crs = QgsCoordinateReferenceSystem(value)

	@property
	def shapefile(self):
		return self.__shapefile

	@property
	def isValid(self):
		if self.shapefile.isValid():
			return True
		return False

	@property
	def hasFeatures(self):
		if self.shapefile.featureCount():
			return True
		return False

	def getFeatures(self, request=None):
		if request is None:
			return [feature for feature in self.shapefile.getFeatures()]
		return [feature for feature in self.shapefile.getFeatures(request)]

	def getGeometries(self):
		return [feature.geometry() for feature in self.getFeatures()]

	@property
	def hasValidCrs(self):
		if self.shapefile.crs() == self.crs:
			return True
		return False

	@property
	def fields(self):
		return [field for field in self.shapefile.pendingFields()]

	def hasField(self, name):
		if name in [field.name() for field in self.fields]:
			return True
		return False


	@property
	def fieldsStructure(self):
		return self.__fieldsStructure
		
		
	def validateFields(self):
		actualFields = {field.name() for field in self.fields}
		minimumFields = {structure["name"] for structure in self.fieldsStructure}
		for field in minimumFields:
			if not field in actualFields:
				yield {"field": field, "minField": True, "check": False}
			else:
				for qfield in self.fields:
					if qfield.name() == field:
						validator = FieldValidator(qfield, self.fieldsStructure)
						yield validator.check()
		"""    
		for field in self.fields:
			validator = FieldValidator(field, self.fieldsStructure)
			yield validator.check()
		"""

	def validateGeometries(self):
		for feature in self.getFeatures():
			validator = GeometryValidator(feature)
			yield validator.check()

	def validateTopology(self, other=None):
		if other is None:
			featuresCombo = combinations(self.getFeatures(), 2)
			for combo in featuresCombo:
				feature, otherFeature = combo
				validator = TopologyValidator(feature, otherFeature)
				yield validator.check()
		else:
			for feature in self.getFeatures():
				for otherFeature in other.getFeatures():
					validator = TopologyValidator(feature, otherFeature)
					yield validator.check()

	def validateAttributes(self, fieldsList=None):
		if fieldsList is None:
			#fieldsList = self.fields
			actualFields = {field.name() for field in self.fields} #aggiunto
			minimumFields = {structure["name"] for structure in self.fieldsStructure} #aggiunto
			for field in minimumFields: #aggiunto
				if not field in actualFields: #aggiunto
					yield {"field": field, "minField": True, "check": False} #aggiunto
				else: #aggiunto
					for qfield in self.fields: #aggiunto
						if qfield.name() == field: #aggiunto
							validator = AttributeValidator(qfield, self.fieldsStructure, self.fields, self.shapefile) #aggiunto
							if validator.hasDomain(): #aggiunto
								yield validator.check() #aggiunto
		#for field in fieldsList:
		   # validator = AttributeValidator(field, self.fieldsStructure, self.fields, self.shapefile)
		   # if validator.hasDomain():
			 #   yield validator.check()


class FieldValidator(object):

	def __init__(self, field, fieldsStructure):
		super(FieldValidator, self).__init__()

		self.__field = field
		self.__fieldsStructure = fieldsStructure

	@property
	def fieldName(self):
		return self.__field.name()

	@property
	def fieldType(self):
		return self.__field.typeName()

	@property
	def fieldLength(self):
		return self.__field.length()

	@property
	def fieldPrecision(self):
		return self.__field.precision()

	@property
	def fieldsStructure(self):
		return self.__fieldsStructure

	def getStructure(self):
		for structure in self.fieldsStructure:
			if structure["name"] == self.fieldName:
				return structure

	def check(self):
		minimumFields = {structure["name"] for structure in self.fieldsStructure}
		
		properties = {}
		structure = self.getStructure()
		if structure:
			properties["name"] = False if self.fieldName != structure["name"] else True
			properties["type"] = False if self.fieldType != structure["type"] else True
			#properties["length"] = False if self.fieldLength != structure["length"] else True
			#properties["precision"] = False if self.fieldPrecision != structure["precision"] else True
		else:
			properties["name"] = False
			expectedName = {"name": [field["name"] for field in self.fieldsStructure]}
		check = all(properties.values())
		return {"field": self.fieldName, "check": check, "properties": properties, "expected": structure if properties["name"] else expectedName}


class GeometryValidator(object):

	def __init__(self, feature):
		super(GeometryValidator, self).__init__()

		self.__geometry = feature.geometry()
		self.__id = feature.id()

	@property
	def geometry(self):
		return self.__geometry

	@property
	def id(self):
		return self.__id

	def check(self):
		check = True
		if self.geometry is None: #modificato Gianluca 
			check = False #modificato Gianluca 
			errs = ("La feature con id " + str(self.id) + " non ha geometria",) #modificato Gianluca 
		else:	 #modificato Gianluca 
			errs = self.geometry.validateGeometry()
			if errs:
				check = False
				errs = [err for err in errs if err.hasWhere()]
		return {"check": check, "errors": errs}


class TopologyValidator(object):

	def __init__(self, feature, otherFeature):
		super(TopologyValidator, self).__init__()

		self.__geometry = feature.geometry()
		self.__otherGeometry = otherFeature.geometry()

	@property
	def geometry(self):
		return self.__geometry

	@property
	def otherGeometry(self):
		return self.__otherGeometry

	def isOverlapped(self):
		return self.geometry.intersects(self.otherGeometry)

	def overlappedGeometry(self):
		return self.geometry.intersection(self.otherGeometry)
		
	def isGeometryValid(self):
		if self.geometry is None or not self.geometry.isGeosValid():
			return False
		return True
		
	def isOtherGeometryValid(self):
		if self.otherGeometry is None or not self.otherGeometry.isGeosValid():
			return False
		return True
		
	def check(self):
		check = True
		intersection = None
		if self.isGeometryValid() and self.isOtherGeometryValid():
			if self.isOverlapped():
				check = False
				intersection = self.overlappedGeometry()
		return {"check": check, "intersection": intersection}


class AttributeValidator(object):

	def __init__(self, field, fieldsStructure, fieldsList, shapefile):
		super(AttributeValidator, self).__init__()

		self.__field = field
		self.__fieldsList = fieldsList
		self.__shapefile = shapefile
		for field in fieldsStructure:
			if field["name"] == self.fieldName:
				self.__fieldStructure = field

	@property
	def shapefile(self):
		return self.__shapefile

	@property
	def field(self):
		return self.__field

	@property
	def fields(self):
		return self.__fieldsList

	@property
	def fieldName(self):
		return self.field.name()

	@property
	def fieldStructure(self):
		return self.__fieldStructure

	def hasDomain(self):
		if "domain" in self.fieldStructure.keys():
			return True
		return False

	@property
	def domain(self):
		if self.hasDomain():
			return self.fieldStructure["domain"]

	def check(self):
		check = False
		errs = []
		#request = QgsFeatureRequest()
		#request.setSubsetOfAttributes([self.fieldName], self.shapefile.pendingFields())
		#request.setFlags(QgsFeatureRequest.NoGeometry)
		for row in self.shapefile.getFeatures():
			if row[self.fieldName] not in self.domain:
				errs.append(int(row.id()) + 1)
		if not errs:
			check = True
		return {"field": self.fieldName, "check": check, "featuresId": errs}
